# ******************************************************************************
# Copyright (c) 2019 Analog Devices, Inc.  All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# - Redistributions of source code must retain the above copyright notice, this
#  list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice,
#  this list of conditions and the following disclaimer in the documentation
#  and/or other materials provided with the distribution.
# - Modified versions of the software must be conspicuously marked as such.
# - This software is licensed solely and exclusively for use with
#  processors/products manufactured by or for Analog Devices, Inc.
# - This software may not be combined or merged with other code in any manner
#  that would cause the software to become subject to terms and conditions
#  which differ from those listed here.
# - Neither the name of Analog Devices, Inc. nor the names of its contributors
#  may be used to endorse or promote products derived from this software
#  without specific prior written permission.
# - The use of this software may or may not infringe the patent rights of one
#  or more patent holders.  This license does not release you from the
#  requirement that you obtain separate licenses from these patent holders to
#  use this software.
#
# THIS SOFTWARE IS PROVIDED BY ANALOG DEVICES, INC. AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# NONINFRINGEMENT, TITLE, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL ANALOG DEVICES, INC. OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, PUNITIVE OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, DAMAGES ARISING OUT OF
# CLAIMS OF INTELLECTUAL PROPERTY RIGHTS INFRINGEMENT; PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# ******************************************************************************


import logging
from typing import List, Dict

from .csv_logging import CSVLogger
from .common_stream import CommonStream
from ..core.enums.dcb_enums import DCBCommand
from ..core.packets.command_packet import CommandPacket
from ..core.enums.adxl_enums import ADXLDevice, ADXLCommand
from ..core.packets.stream_data_packets import ADXLDataPacket
from ..core.packets.common_packets import DecimationFactorPacket
from ..core.enums.common_enums import Application, Stream, CommonCommand
from ..core.packets.adxl_packets import ADXLConfigPacket, ADXLRegisterWritePacket, ADXLRegisterReadPacket, \
    ADXLDCBPacket, ADXLDCFGPacket, ADXLDCBCommandPacket

logger = logging.getLogger(__name__)


class ADXLApplication(CommonStream):
    """
    ADXL Application class.

    .. code-block:: python3
        :emphasize-lines: 4

        from adi_study_watch import SDK

        sdk = SDK("COM4")
        application = sdk.get_adxl_application()

    """
    DEVICE_362 = ADXLDevice.DEVICE_362

    def __init__(self, packet_manager):
        super().__init__(Application.ADXL, Stream.ADXL, packet_manager)
        self._dcb_size = 25

    @staticmethod
    def get_supported_streams() -> List[Stream]:
        """
        List all supported streams.

        :return: Array of stream ID enums.
        :rtype: List[Stream]

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_adpd_application()
            x = application.get_supported_streams()
            print(x)
            # [<Stream.ADPD1: ['0xC2', '0x11']>, ... , <Stream.ADPD12: ['0xC2', '0x1D']>]
        """
        return [Stream.ADXL]

    def delete_device_configuration_block(self) -> Dict:
        """
        Deletes Adxl Device configuration block.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_adxl_application()
            application.delete_device_configuration_block()
        """
        request_packet = ADXLDCBCommandPacket(self._destination, DCBCommand.ERASE_CONFIG_REQ)
        response_packet = ADXLDCBCommandPacket(self._destination, DCBCommand.ERASE_CONFIG_RES)
        return self._send_packet(request_packet, response_packet)

    def get_decimation_factor(self) -> Dict:
        """
        Returns stream decimation factor.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_adxl_application()
            x = application.get_decimation_factor()
            print(x["payload"]["decimation_factor"])
            # 1

        """
        request_packet = DecimationFactorPacket(self._destination, CommonCommand.GET_STREAM_DEC_FACTOR_REQ)
        request_packet.set_payload("stream_address", self._stream)
        response_packet = DecimationFactorPacket(self._destination, CommonCommand.GET_STREAM_DEC_FACTOR_RES)
        return self._send_packet(request_packet, response_packet)

    def get_device_configuration(self) -> Dict:
        """
        Returns device configuration data.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_adxl_application()
            x = application.get_device_configuration()
            print(x["payload"]["dcfg_data"])
            # [['0x9', '0x97'], ['0x7', '0x8FFF'], ['0xB', '0x2F6'], ... ]
        """
        request_packet = CommandPacket(self._destination, CommonCommand.GET_DCFG_REQ)
        response_packet = ADXLDCFGPacket(self._destination, CommonCommand.GET_DCFG_RES)
        return self._send_packet(request_packet, response_packet)

    @staticmethod
    def get_supported_devices() -> List[ADXLDevice]:
        """
        List all supported device ID for adxl.

        :return: Array of device ID enums.
        :rtype: List

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_adxl_application()
            x = application.get_supported_devices()
            print(x)
            # [<ADXLDevice.DEVICE_362: ['0x6A', '0x1']>]
        """
        return [ADXLApplication.DEVICE_362]

    def load_configuration(self, device_id: ADXLDevice) -> Dict:
        """
        Loads specified device id configuration.

        :param device_id: Device ID to load, use get_supported_devices() to list all supported devices.
        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5,8

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_adxl_application()
            x = application.get_supported_devices()
            print(x)
            # [<ADXLDevice.DEVICE_362: ['0x6A', '0x1']>]
            x = application.load_configuration(application.DEVICE_362)
            print(x["payload"]["device_id"])
            # <ADXLDevice.DEVICE_362: ['0x6A', '0x1']>

        """
        request_packet = ADXLConfigPacket(self._destination, ADXLCommand.LOAD_CONFIG_REQ)
        request_packet.set_payload("device_id", device_id)
        response_packet = ADXLConfigPacket(self._destination, ADXLCommand.LOAD_CONFIG_RES)
        return self._send_packet(request_packet, response_packet)

    def read_device_configuration_block(self) -> Dict:
        """
        Returns entire device configuration block.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_adxl_application()
            x = application.read_device_configuration_block()
            print(x["payload"]["data"])
            # []

        """
        request_packet = ADXLDCBCommandPacket(self._destination, DCBCommand.READ_CONFIG_REQ)
        response_packet = ADXLDCBPacket(self._destination, DCBCommand.READ_CONFIG_RES)
        return self._send_packet(request_packet, response_packet)

    def read_register(self, addresses: List[int]) -> Dict:
        """
        Reads the register values of specified addresses. This function takes a list of addresses to read,
        and returns a response packet as dictionary containing addresses and values.

        :param addresses: List of register addresses to read.
        :type addresses: List[int]
        :return: A response packet as dictionary
        :rtype: Dict

        .. list-table::
           :widths: 50 50
           :header-rows: 1

           * - Address Lower Limit
             - Address Upper Limit
           * - 0x00
             - 0x2E

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_adxl_application()
            x = application.read_register([0x15, 0x20, 0x2E])
            print(x["payload"]["data"])
            # [['0x15', '0x0'], ['0x20', '0x0'], ['0x2E', '0x0']]
        """
        data = [[address, 0] for address in addresses]
        request_packet = ADXLRegisterReadPacket(self._destination, CommonCommand.REGISTER_READ_REQ)
        request_packet.set_payload("size", len(data))
        request_packet.set_payload("data", data)
        response_packet = ADXLRegisterReadPacket(self._destination, CommonCommand.REGISTER_READ_RES)
        return self._send_packet(request_packet, response_packet)

    def set_decimation_factor(self, decimation_factor: int) -> Dict:
        """
        Sets decimation factor for adxl stream.

        :param decimation_factor: decimation factor for stream
        :type decimation_factor: int
        :return: A response packet as dictionary
        :rtype: Dict

        .. list-table::
           :widths: 50 50
           :header-rows: 1

           * - Decimation Lower Limit
             - Decimation Upper Limit
           * - 0x01
             - 0x05

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_adxl_application()
            x = application.set_decimation_factor(2)
            print(x["payload"]["decimation_factor"])
            # 2

        """
        request_packet = DecimationFactorPacket(self._destination, CommonCommand.SET_STREAM_DEC_FACTOR_REQ)
        request_packet.set_payload("stream_address", self._stream)
        request_packet.set_payload("decimation_factor", decimation_factor)
        response_packet = DecimationFactorPacket(self._destination, CommonCommand.SET_STREAM_DEC_FACTOR_RES)
        return self._send_packet(request_packet, response_packet)

    def write_device_configuration_block(self, addresses_values: List[List[int]]) -> Dict:
        """
        Writes the device configuration block values of specified addresses.
        This function takes a list of addresses and values to write, and returns a response packet as
        dictionary containing addresses and values.

        :param addresses_values: List of addresses and values to write.
        :type addresses_values: List[List[int]]
        :return: A response packet as dictionary.
        :rtype: Dict

        .. list-table::
           :widths: 50 50
           :header-rows: 1

           * - Address Lower Limit
             - Address Upper Limit
           * - 0x20
             - 0x2E

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_adxl_application()
            x = application.write_device_configuration_block([[0x20, 2], [0x21, 0x1]])
            print(x["payload"]["size"])
            # 2

        """
        request_packet = ADXLDCBPacket(self._destination, DCBCommand.WRITE_CONFIG_REQ)
        request_packet.set_payload("size", len(addresses_values))
        request_packet.set_payload("data", addresses_values)
        response_packet = ADXLDCBCommandPacket(self._destination, DCBCommand.WRITE_CONFIG_RES)
        return self._send_packet(request_packet, response_packet)

    def write_device_configuration_block_from_file(self, filename: str) -> Dict:
        """
        Writes the device configuration block values of specified addresses from file.

        :param filename: dcb filename
        :return: A response packet as dictionary.
        :rtype: Dict

        .. list-table::
           :widths: 50 50
           :header-rows: 1

           * - Address Lower Limit
             - Address Upper Limit
           * - 0x20
             - 0x2E

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_adxl_application()
            application.write_device_configuration_block_from_file("adxl_dcb.dcfg")

        """
        result = self.device_configuration_file_to_list(filename)
        if result:
            return self.write_device_configuration_block(result)

    def write_register(self, addresses_values: List[List[int]]) -> Dict:
        """
        Writes the register values of specified addresses. This function takes a list of addresses and values to write,
        and returns a response packet as dictionary containing addresses and values.

        :param addresses_values: List of register addresses and values to write.
        :type addresses_values: List[List[int]]
        :return: A response packet as dictionary.
        :rtype: Dict

        .. list-table::
           :widths: 50 50
           :header-rows: 1

           * - Address Lower Limit
             - Address Upper Limit
           * - 0x20
             - 0x2E

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_adxl_application()
            x = application.write_register([[0x20, 0x1], [0x21, 0x2], [0x2E, 0x3]])
            print(x["payload"]["data"])
            # [['0x20', '0x1'], ['0x21', '0x2'], ['0x2E', '0x3']]

        """
        request_packet = ADXLRegisterWritePacket(self._destination, CommonCommand.REGISTER_WRITE_REQ)
        request_packet.set_payload("size", len(addresses_values))
        request_packet.set_payload("data", addresses_values)
        response_packet = ADXLRegisterWritePacket(self._destination, CommonCommand.REGISTER_WRITE_RES)
        return self._send_packet(request_packet, response_packet)

    def _callback_data(self, packet, packet_id, callback_function=None, args=None):
        """
        Process and returns the data back to user's callback function.
        """
        self._callback_data_helper(packet, ADXLDataPacket())

    def enable_csv_logging(self, filename: str, header: List = None) -> None:
        """
        Start logging stream data into CSV.

        :param filename: Name of the CSV file.
        :param header: Header list of the CSV file.
        :return: None

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_adxl_application()
            x = application.enable_csv_logging("adxl.csv")
        """
        if header is None:
            header = ["Timestamp", "X", "Y", "Z"]
        self._csv_logger[Stream.ADXL] = CSVLogger(filename, header)

    def disable_csv_logging(self) -> None:
        """
        Stops logging stream data into CSV.

        :return: None

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_adxl_application()
            x = application.disable_csv_logging()
        """
        if self._csv_logger.get(Stream.ADXL):
            self._csv_logger[Stream.ADXL].stop_logging()
        self._csv_logger[Stream.ADXL] = None



# ******************************************************************************
# Copyright (c) 2019 Analog Devices, Inc.  All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# - Redistributions of source code must retain the above copyright notice, this
#  list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice,
#  this list of conditions and the following disclaimer in the documentation
#  and/or other materials provided with the distribution.
# - Modified versions of the software must be conspicuously marked as such.
# - This software is licensed solely and exclusively for use with
#  processors/products manufactured by or for Analog Devices, Inc.
# - This software may not be combined or merged with other code in any manner
#  that would cause the software to become subject to terms and conditions
#  which differ from those listed here.
# - Neither the name of Analog Devices, Inc. nor the names of its contributors
#  may be used to endorse or promote products derived from this software
#  without specific prior written permission.
# - The use of this software may or may not infringe the patent rights of one
#  or more patent holders.  This license does not release you from the
#  requirement that you obtain separate licenses from these patent holders to
#  use this software.
#
# THIS SOFTWARE IS PROVIDED BY ANALOG DEVICES, INC. AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# NONINFRINGEMENT, TITLE, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL ANALOG DEVICES, INC. OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, PUNITIVE OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, DAMAGES ARISING OUT OF
# CLAIMS OF INTELLECTUAL PROPERTY RIGHTS INFRINGEMENT; PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# ******************************************************************************

from unittest import TestCase

from sdk_pure_python.adi_study_watch import SDK
from sdk_pure_python.adi_study_watch.core.enums.fs_enums import FSStatus
from sdk_pure_python.adi_study_watch.core.enums.pm_enums import PMStatus
from sdk_pure_python.adi_study_watch.core.enums.common_enums import CommonStatus


def callback(data):
    print(data)


class TestTemperatureApplication(TestCase):

    @classmethod
    def setUpClass(cls):
        cls.application = SDK("COM4", debug=True).get_test_application(callback, callback)

    def test_ping(self):
        x = self.application.ping(3, 70)
        assert (x[0]["payload"]["status"] == CommonStatus.OK)

    # def test_set_battery_threshold(self):
    #     x = self.application.set_battery_threshold(20, 5)
    #     print(x)
    #     assert (x["payload"]["status"] == PMStatus.OK)

    def test_set_power_mode(self):
        x = self.application.set_power_mode(self.application.POWER_MODE_ACTIVE)
        assert (x["payload"]["status"] == PMStatus.OK)

    def test_get_apps_health_status(self):
        x = self.application.get_apps_health_status()
        assert (x["payload"]["status"] == CommonStatus.OK)

    def test_flash_reset(self):
        x = self.application.flash_reset()
        assert (x["payload"]["status"] == PMStatus.OK)

    def test_get_states(self):
        x = self.application.get_ppg_states()
        print(x)

    def test_electrode_switch(self):
        x = self.application.disable_electrode_switch(self.application.SWITCH_ADPD4000)
        assert (x["payload"]["status"] == PMStatus.OK)
        x = self.application.enable_electrode_switch(self.application.SWITCH_ADPD4000)
        assert (x["payload"]["status"] == PMStatus.OK)

    def test_ldo(self):
        x = self.application.disable_ldo(self.application.LDO_OPTICAL)
        assert (x["payload"]["status"] == PMStatus.OK)
        x = self.application.enable_ldo(self.application.LDO_OPTICAL)
        assert (x["payload"]["status"] == PMStatus.OK)

    def test_get_bad_blocks(self):
        x = self.application.get_bad_blocks()
        # assert (x["payload"]["status"] == CommonStatus.OK)

    def test_get_debug_info(self):
        x = self.application.get_debug_info()
        # assert (x["payload"]["status"] == CommonStatus.OK)

    def test_get_stream_debug_info(self):
        x = self.application.get_stream_debug_info(self.application.STREAM_ADXL)
        assert (x["payload"]["status"] == FSStatus.OK)

    def test_fds_status(self):
        self.application.get_fds_status()

    def test_get_agc_info(self):
        x = self.application.get_agc_info(self.application.LED_GREEN)
        assert (x["payload"]["led_index"] == self.application.LED_GREEN)

    def test_adxl_self_test(self):
        x = self.application.adxl_self_test()
        assert (x["payload"]["status"] == CommonStatus.OK)

    def test_file_read_test(self):
        x = self.application.file_read_test("03124671.LOG")
        # assert (x["payload"]["status"] == FSStatus.OK)

    def test_page_read_test(self):
        x = self.application.page_read_test(517, 5)
        assert (x["payload"]["status"] == FSStatus.OK)

    def test_get_file_info(self):
        x = self.application.get_file_info(3)
        # assert (x["payload"]["status"] == FSStatus.OK)

    def test_get_fds_status(self):
        x = self.application.get_fds_status()
        print(x)

    def test_read_device_configuration_block_info(self):
        x = self.application.read_device_configuration_block_info()
        print(x)

    def test_command1(self):
        x = self.application.test_command1(5)
        print(x)



# ******************************************************************************
# Copyright (c) 2023 Analog Devices, Inc.  All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# - Redistributions of source code must retain the above copyright notice, this
#  list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice,
#  this list of conditions and the following disclaimer in the documentation
#  and/or other materials provided with the distribution.
# - Modified versions of the software must be conspicuously marked as such.
# - This software is licensed solely and exclusively for use with
#  processors/products manufactured by or for Analog Devices, Inc.
# - This software may not be combined or merged with other code in any manner
#  that would cause the software to become subject to terms and conditions
#  which differ from those listed here.
# - Neither the name of Analog Devices, Inc. nor the names of its contributors
#  may be used to endorse or promote products derived from this software
#  without specific prior written permission.
# - The use of this software may or may not infringe the patent rights of one
#  or more patent holders.  This license does not release you from the
#  requirement that you obtain separate licenses from these patent holders to
#  use this software.
#
# THIS SOFTWARE IS PROVIDED BY ANALOG DEVICES, INC. AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# NONINFRINGEMENT, TITLE, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL ANALOG DEVICES, INC. OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, PUNITIVE OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, DAMAGES ARISING OUT OF
# CLAIMS OF INTELLECTUAL PROPERTY RIGHTS INFRINGEMENT; PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# ******************************************************************************

"""
    Implements ConnectionInterface for NRF BLE UART service using host device built in BLE
"""
import logging
import sys
import threading
import time
from asyncio import Task
from typing import Optional

from .ipc_tools import BaseThread, FifoQueue
from .connection_interface import ConnectionInterface
from bleak import BleakClient, BleakGATTCharacteristic
import asyncio

logger = logging.getLogger(__name__)

logi = logger.info
logw = logger.warning
loge = logger.error


class NrfUartBleConnection(ConnectionInterface, BaseThread):
    """
        Non-blocking write/read ble  communication
        interface implementation.

        There are two FIFO queues, one for TX and another for RX.
        When write is called, data are pushed to write (TX) queue,
        write thread will peek and send.

        When the data are received, ble characteristic callback
        will push the data to a read (RX) queue, and read() function
        will return them

    """

    # NRF defined constants
    READ_DESCRIPTOR_UUID: str = "6e400001-b5a3-f393-e0a9-e50e24dcca9e"
    WRITE_CHARACTERISTIC_UUID: str = "6e400002-b5a3-f393-e0a9-e50e24dcca9e"
    READ_CHARACTERISTIC_UUID: str = "6e400003-b5a3-f393-e0a9-e50e24dcca9e"

    RX_QUEUE_MAX_SIZE = 100
    TX_QUEUE_MAX_SIZE = 100

    def __init__(self, ble_address: str, *args, **kwargs):
        BaseThread.__init__(self)
        ConnectionInterface.__init__(self, *args, **kwargs)
        self._is_connected: threading.Event = threading.Event()
        self.ble_address: str = ble_address
        self.write_queue: FifoQueue = FifoQueue(maxsize=NrfUartBleConnection.TX_QUEUE_MAX_SIZE)
        self.read_queue: FifoQueue = FifoQueue(maxsize=NrfUartBleConnection.RX_QUEUE_MAX_SIZE)
        self.read_task: Optional[Task] = None
        self.async_loop = None
        self._is_opened: threading.Event = threading.Event()

    def read(self, size=1) -> bytes:
        """
            Read data if any, otherwise immediately return empty. Nor blocking mode
        :param size: number of bytes to read, default 1
        :return: bytes read or empty

        """
        _d = self.read_queue.pop_all(count=size, wait_all=False)
        if _d is None:
            return b''
        return bytes(_d)

    def write(self, data) -> int:
        """
            Write data to BLE, return count written. Non-blocking mode
        :param data: data to write to BLE
        :return: number of bytes written to internal queue, not actually sent
        """
        self.write_queue.push_all(data)
        return len(data)

    def close(self) -> None:
        """
            Stop internal processes and close connection
        :return: none
        """
        logi('NrfUartBleConnection.close::>> CLOSE ')
        if not self._is_opened.is_set():
            logw('NrfUartBleConnection.close::>> Already closed')
            return

        self._is_opened.clear()
        self._is_connected.clear()
        self.stop()

        try:
            logi(f'NrfUartBleConnection.close::>> Running loop {asyncio.get_running_loop()}')
            for task in asyncio.all_tasks():
                logi(f'NrfUartBleConnection.close::>> Cancel task {task}')
                task.cancel()
            logi(f'NrfUartBleConnection.close::>> Running loop {asyncio.get_running_loop()}')
        except Exception as e:
            logw(f'NrfUartBleConnection.close::>> Close exception: {e}')
        logi("NrfUartBleConnection.close::>> Connection closed!")

    def handle_disconnect(self, client: BleakClient):
        """
            Bleak Disconnect callback

        """
        logi(f"NrfUartBleConnection.handle_disconnect::>> Device was disconnected, goodbye: {client}")
        self.close()

    def handle_rx(self, _: BleakGATTCharacteristic, data: bytearray):
        """
            Bleak RX callback
        """
        self.read_queue.push_all(data)

    async def ble_service_task(self):
        """
            BLE asyncio communication task that connects, and waits for messages in TX queue
        :return: none
        """

        logi(f'NrfUartBleConnection.ble_service_task::>> Connecting to device {self.ble_address}')
        try:
            async with BleakClient(self.ble_address, disconnected_callback=self.handle_disconnect) as client:

                await client.start_notify(NrfUartBleConnection.READ_CHARACTERISTIC_UUID, self.handle_rx)

                nus = client.services.get_service(NrfUartBleConnection.READ_DESCRIPTOR_UUID)
                tx_char = nus.get_characteristic(NrfUartBleConnection.WRITE_CHARACTERISTIC_UUID)

                self._is_connected.set()
                logi(f'NrfUartBleConnection.ble_service_task::>> Connected to client {client}')
                while self.is_running() and client.is_connected:
                    if not self.write_queue.empty():
                        _d = self.write_queue.pop_all(count=sys.maxsize, wait_all=False)
                        await client.write_gatt_char(tx_char, bytearray(_d))
                    else:
                        await asyncio.sleep(0.01)
                if client.is_connected:
                    logi(f'NrfUartBleConnection.ble_service_task::>> Stop communication with {client}, '
                         f'disconnecting ...')
                    await client.disconnect()

                self._is_connected.clear()
                logi(f'NrfUartBleConnection.ble_service_task::>> Communication with {client} stopped, close connection')
                self.close()

        except Exception as e:
            loge(f'NrfUartBleConnection.ble_service_task::>> ble_service_task exception: - {e}')
            self.close()

        logi('NrfUartBleConnection.ble_service_task::>> BLE Task stopped')

    def open(self, timeout: int = 0) -> bool:
        """
            Open connection with optional timeout
        :param timeout: time out to wait before fail. If zero, will wait forever
        :return: True if connection is successful
        :note: By default, this function will block forever
        """
        self.start(start_delay=0)
        logi('NrfUartBleConnection.open::>> Open ...')
        try:
            start_time: int = int(time.time())
            while self.is_running() and not self.is_connected():
                time.sleep(1)
                if 0 < timeout <= int(time.time()) - start_time:
                    break
            self._is_opened.set()

            return self.is_connected()
        except Exception as ex:
            loge(f'NrfUartBleConnection.open::>> Failed to connect {ex}')

        return False

    def loop(self):
        """
            Wraps asyncio call
        :return: none
        """
        try:
            self.async_loop = asyncio.new_event_loop()
            self.async_loop.run_until_complete(self.ble_service_task())
        except asyncio.CancelledError as e:
            loge(e)

        self.close()

    def cancel_read(self) -> None:
        """
            Not blocking when read
        :return: none
        """
        pass

    def is_connected(self) -> bool:
        """
            Check if device is connected (communication channel is open)
            :return: True or false
        """
        return self._is_connected.is_set()



# ******************************************************************************
# Copyright (c) 2023 Analog Devices, Inc.  All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# - Redistributions of source code must retain the above copyright notice, this
#  list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice,
#  this list of conditions and the following disclaimer in the documentation
#  and/or other materials provided with the distribution.
# - Modified versions of the software must be conspicuously marked as such.
# - This software is licensed solely and exclusively for use with
#  processors/products manufactured by or for Analog Devices, Inc.
# - This software may not be combined or merged with other code in any manner
#  that would cause the software to become subject to terms and conditions
#  which differ from those listed here.
# - Neither the name of Analog Devices, Inc. nor the names of its contributors
#  may be used to endorse or promote products derived from this software
#  without specific prior written permission.
# - The use of this software may or may not infringe the patent rights of one
#  or more patent holders.  This license does not release you from the
#  requirement that you obtain separate licenses from these patent holders to
#  use this software.
#
# THIS SOFTWARE IS PROVIDED BY ANALOG DEVICES, INC. AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# NONINFRINGEMENT, TITLE, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL ANALOG DEVICES, INC. OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, PUNITIVE OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, DAMAGES ARISING OUT OF
# CLAIMS OF INTELLECTUAL PROPERTY RIGHTS INFRINGEMENT; PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# ******************************************************************************

from enum import Enum

from .. import utils


class SessionManagerStatus(Enum):
    """
    SessionManager Status Enum
    """
    LOWEST = [0x40]
    OK = [0x41]
    ERR_ARGS = [0x42]
    ERR_NOT_CHKD = [0xFF]

    def __repr__(self):
        return "<%s.%s: %r>" % (self.__class__.__name__, self._name_, utils.convert_int_array_to_hex(self._value_))


class SessionManagerCommand(Enum):
    """
    SessionManager Command Enums
    """
    LOWEST = [0x40]
    SET_STATE_REQ = [0x41]
    SET_STATE_RES = [0x42]
    GET_STATE_REQ = [0x43]
    GET_STATE_RES = [0x44]

    def __repr__(self):
        return "<%s.%s: %r>" % (self.__class__.__name__, self._name_, utils.convert_int_array_to_hex(self._value_))


class SessionManagerState(Enum):
    """
    SessionManager States Enums
    """
    STANDBY = [0x0]
    CONFIGURED = [0x1]
    LIVESTREAM = [0x2]
    RECORD = [0x3]
    RECORD_START_LOG = [0x4]
    RECORD_STOP_LOG = [0x5]
    END_LIVESTREAM = [0x6]
    END_RECORD = [0x7]

    def __repr__(self):
        return "<%s.%s: %r>" % (self.__class__.__name__, self._name_, utils.convert_int_array_to_hex(self._value_))


class SessionManagerEvent(Enum):
    """
    SessionManager Events Enums
    """
    INVALID = [0x0]
    NAV_BUTTON_RESET = [0x1]
    BATTERY_CRITICAL = [0x2]
    BATTERY_DRAINED = [0x3]
    BLE_DISCONNECT_UNEXPECTED = [0x4]
    BLE_DISCONNECT_SESSION_TERMINATED = [0x5]
    RTC_TIMER_INTERRUPT = [0x6]
    BLE_ADV_TIMEOUT = [0x7]
    USB_DISCONNECT_UNEXPECTED = [0x8]
    BATTERY_FULL = [0x9]
    FINISH_LOG_TRANSFER = [0xA]
    SYS_RST_M2M2_COMMAND = [0xB]
    SYS_HW_RST_M2M2_COMMAND = [0xC]

    def __repr__(self):
        return "<%s.%s: %r>" % (self.__class__.__name__, self._name_, utils.convert_int_array_to_hex(self._value_))


class WatchResetReason(Enum):
    """
    Watch Reset Reasons Enums
    """
    INVALID = [0x0]
    RST_PIN_RESET = [0x1]
    NRF_WDT_RESET = [0x2]
    SOFT_RESET = [0x4]
    CPU_LOCKUP = [0x8]

    def __repr__(self):
        return "<%s.%s: %r>" % (self.__class__.__name__, self._name_, utils.convert_int_array_to_hex(self._value_))



from enum import Enum, unique

from .. import utils


@unique
class LTMode4Command(Enum):
    """
    LTMode4 Command Enum
    """
    LOWEST = [0x40]
    SET_STATE_REQ = [0x42]
    SET_STATE_RES = [0x43]
    GET_STATE_REQ = [0x44]
    GET_STATE_RES = [0x45]
    CLEAR_PREV_ST_EVT_REQ = [0x48]
    CLEAR_PREV_ST_EVT_RES = [0x49]
    GET_PREV_ST_EVT_REQ = [0x4A]
    GET_PREV_ST_EVT_RES = [0x4B]

    def __repr__(self):
        return "<%s.%s: %r>" % (self.__class__.__name__, self._name_, utils.convert_int_array_to_hex(self._value_))


@unique
class LTMode4Status(Enum):
    """
    LTMode4 status Enum
    """
    LOWEST = [0x40]
    OK = [0x41]
    ERR_ARGS = [0x42]
    ERR_NOT_CHKD = [0xFF]

    def __repr__(self):
        return "<%s.%s: %r>" % (self.__class__.__name__, self._name_, utils.convert_int_array_to_hex(self._value_))


@unique
class LTMode4State(Enum):
    """
    LTMode4 status Enum
    """
    STANDBY = [0x0]
    CONFIGURED = [0x1]
    LOG_ON = [0x2]
    LOG_OFF = [0x3]
    FS_MEMORY_FULL = [0x4]
    FS_MAX_FILE_COUNT = [0x5]
    SHIPMENT_MODE = [0x6]
    LOG_OFF_TOOL_CONNECTED = [0x7]
    LOG_DOWNLOAD = [0x8]

    def __repr__(self):
        return "<%s.%s: %r>" % (self.__class__.__name__, self._name_, utils.convert_int_array_to_hex(self._value_))


@unique
class LTMode4Event(Enum):
    """
    LTMode4 event Enum
    """
    NO_EVENT = [0x0]
    WATCH_OFF_CRADLE_NAV_BUTTON_RESET = [0x1]
    WATCH_ON_CRADLE_NAV_BUTTON_RESET = [0x2]
    WATCH_CHARGING_STARTED = [0x3]
    WATCH_CHARGING_STOPPED_MEMORY_FULL = [0x4]
    WATCH_CHARGING_STOPPED_MAXIMUM_FILE_COUNT = [0x5]
    WATCH_CHARGING_STOPPED_MEMORY_OK = [0x6]
    FS_MEMORY_FULL = [0x7]
    FS_MAX_FILE_COUNT = [0x8]
    GET_VERSION_M2M2_CMD = [0x9]
    SHIPMENT_M2M2_CMD = [0xA]
    DELETE_LT_CONFIG_M2M2_CMD = [0xB]
    SET_STATE_M2M2_CMD = [0xC]
    WATCH_BOOT_NO_SOFT_RESET = [0xD]
    SYS_RST_M2M2_COMMAND = [0xE]

    def __repr__(self):
        return "<%s.%s: %r>" % (self.__class__.__name__, self._name_, utils.convert_int_array_to_hex(self._value_))


@unique
class LTMode4WatchResetReason(Enum):
    """
    LTMode4 Watch Reset Reason Enum
    """
    INVALID = [0x0]
    RST_PIN_RESET = [0x1]
    NRF_WDT_RESET = [0x2]
    SOFT_RESET = [0x4]
    CPU_LOCKUP = [0x8]

    def __repr__(self):
        return "<%s.%s: %r>" % (self.__class__.__name__, self._name_, utils.convert_int_array_to_hex(self._value_))



# ******************************************************************************
# Copyright (c) 2019 Analog Devices, Inc.  All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# - Redistributions of source code must retain the above copyright notice, this
#  list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice,
#  this list of conditions and the following disclaimer in the documentation
#  and/or other materials provided with the distribution.
# - Modified versions of the software must be conspicuously marked as such.
# - This software is licensed solely and exclusively for use with
#  processors/products manufactured by or for Analog Devices, Inc.
# - This software may not be combined or merged with other code in any manner
#  that would cause the software to become subject to terms and conditions
#  which differ from those listed here.
# - Neither the name of Analog Devices, Inc. nor the names of its contributors
#  may be used to endorse or promote products derived from this software
#  without specific prior written permission.
# - The use of this software may or may not infringe the patent rights of one
#  or more patent holders.  This license does not release you from the
#  requirement that you obtain separate licenses from these patent holders to
#  use this software.
#
# THIS SOFTWARE IS PROVIDED BY ANALOG DEVICES, INC. AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# NONINFRINGEMENT, TITLE, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL ANALOG DEVICES, INC. OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, PUNITIVE OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, DAMAGES ARISING OUT OF
# CLAIMS OF INTELLECTUAL PROPERTY RIGHTS INFRINGEMENT; PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# ******************************************************************************
from ..data_types.array import Array
from ..data_types.enums import Enums
from ..data_types.integer import Int
from .command_packet import CommandPacket
from ..enums.adp5360_enums import BatteryStatus
from ..enums.dcb_enums import DCBConfigBlockIndex


class BatteryInfoPacket(CommandPacket):
    """
    Packet Structure:

    .. code-block::

        {
            'header': {
                'source': <Application.PM: ['0xC5', '0x00']>,
                'destination': <Application.APP_USB: ['0xC7', '0x05']>,
                'length': '0x12',
                'checksum': '0x0'
            },
            'payload': {
                'command': <PMCommand.GET_BAT_INFO_RES: ['0x45']>,
                'status': <PMStatus.OK: ['0x41']>,
                'timestamp': 668880834,
                'battery_status': <BatteryStatus.COMPLETE: ['0x03']>,
                'adp5360_battery_level': 100,
                'custom_battery_level': 100,
                'battery_mv': 4334
            }
        }
    """

    def __init__(self, destination=None, command=None):
        super().__init__(destination, command)
        self._config["payload"]["timestamp"] = Int(4)
        self._config["payload"]["battery_status"] = Enums(1, enum_class=BatteryStatus)
        self._config["payload"]["adp5360_battery_level"] = Int(1)
        self._config["payload"]["custom_battery_level"] = Int(1)
        self._config["payload"]["battery_mv"] = Int(2)


class BatteryThresholdPacket(CommandPacket):
    """
    Packet Structure:

    .. code-block::

        {
            'header': {
                'source': <Application.PM: ['0xC5', '0x00']>,
                'destination': <Application.APP_USB: ['0xC7', '0x05']>,
                'length': '0xA',
                'checksum': '0x0'
            },
            'payload': {
                'command': <PMCommand.SET_BAT_THR_RES: ['0x47']>,
                'status': <PMStatus.OK: ['0x41']>
            }
        }
    """

    def __init__(self, destination=None, command=None):
        super().__init__(destination, command)
        self._config["payload"]["low_level"] = Int(1, value_limit=[0, 100])
        self._config["payload"]["critical_level"] = Int(1, value_limit=[0, 100])
        self._config["payload"]["download_level"] = Int(1)
        self._config["payload"]["voltage_low"] = Int(2)
        self._config["payload"]["voltage_critical"] = Int(2)
        self._config["payload"]["voltage_critical_download"] = Int(2)


class ADP5360RegisterWritePacket(CommandPacket):
    """
    Packet Structure:

    .. code-block::

        {
            'header': {
                'source': <Application.PM: ['0xC5', '0x00']>,
                'destination': <Application.APP_USB: ['0xC7', '0x05']>,
                'length': 15,
                'checksum': 0
            },
            'payload': {
                'command': <CommonCommand.REGISTER_WRITE_RES: ['0x24']>,
                'status': <CommonStatus.OK: ['0x00']>,
                'size': 1,
                'data': [
                    [ '0x0', '0x10' ]
                ]
            }
        }
    """

    def __init__(self, destination=None, command=None):
        super().__init__(destination, command)
        self._config["payload"]["size"] = Int(1)
        self._config["payload"]["data"] = Array(-1, dimension=2,
                                                data_types=[
                                                    Int(2, to_hex=True),
                                                    Int(2, to_hex=True)
                                                ])


class ADP5360RegisterReadPacket(CommandPacket):
    """
    Packet Structure:

    .. code-block::

        {
            'header': {
                'source': <Application.PM: ['0xC5', '0x00']>,
                'destination': <Application.APP_USB: ['0xC7', '0x05']>,
                'length': 15,
                'checksum': 0
            },
            'payload': {
                'command': <CommonCommand.REGISTER_READ_RES: ['0x22']>,
                'status': <CommonStatus.OK: ['0x00']>,
                'size': 1,
                'data': [
                    [ '0x0', '0x10' ]
                ]
            }
        }
    """

    def __init__(self, destination=None, command=None):
        super().__init__(destination, command)
        self._config["payload"]["size"] = Int(1)
        self._config["payload"]["data"] = Array(-1, dimension=2,
                                                data_types=[
                                                    Int(2, value_limit=[0x00, 0x36], to_hex=True),
                                                    Int(2, to_hex=True)
                                                ])


class ADP5360DCBPacket(CommandPacket):
    """
    Packet Structure:

    .. code-block::

        {
            'header': {
                'source': <Application.PM: ['0xC5', '0x00']>,
                'destination': <Application.APP_USB: ['0xC7', '0x05']>,
                'length': 32,
                'checksum': 0
            },
            'payload': {
                'command': <DCBCommand.READ_CONFIG_RES: ['0x98']>,
                'status': <DCBStatus.OK: ['0x97']>,
                'size': 5,
                'data': [
                    [ '0x0', '0x1B58' ],
                    [ '0x1', '0x1388' ],
                    [ '0x2', '0x564' ],
                    [ '0x3', '0x53C' ],
                    [ '0x4', '0x3' ]
                ]
            }
        }
    """

    def __init__(self, destination=None, command=None):
        super().__init__(destination, command)
        self._config["payload"]["dcb_block_index"] = Enums(1, enum_class=DCBConfigBlockIndex,
                                                           default=DCBConfigBlockIndex.ADP5360_BLOCK)
        self._config["payload"]["size"] = Int(1)
        self._config["payload"]["data"] = Array(-1, dimension=2,
                                                data_types=[
                                                    Int(1, to_hex=True),
                                                    Int(3, to_hex=True)
                                                ], reverse_inner_array=True)


class ADP5360DCBCommandPacket(CommandPacket):
    """
    Packet Structure:

    .. code-block::

        {
            'header': {
                'source': <Application.EDA: ['0xC3', '0x02']>,
                'destination': <Application.APP_USB: ['0xC7', '0x05']>,
                'length': 20,
                'checksum': 0
            },
            'payload': {
                'command': <DCBCommand.READ_CONFIG_RES: ['0x98']>,
                'status': <DCBStatus.OK: ['0x97']>,
                'size': 1,
                'data': [
                    [ '0x0', '0x8' ]
                ]
            }
        }
    """

    def __init__(self, destination=None, command=None):
        super().__init__(destination, command)
        self._config["payload"]["dcb_block_index"] = Enums(1, enum_class=DCBConfigBlockIndex,
                                                           default=DCBConfigBlockIndex.ADP5360_BLOCK)



# ******************************************************************************
# Copyright (c) 2019 Analog Devices, Inc.  All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# - Redistributions of source code must retain the above copyright notice, this
#  list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice,
#  this list of conditions and the following disclaimer in the documentation
#  and/or other materials provided with the distribution.
# - Modified versions of the software must be conspicuously marked as such.
# - This software is licensed solely and exclusively for use with
#  processors/products manufactured by or for Analog Devices, Inc.
# - This software may not be combined or merged with other code in any manner
#  that would cause the software to become subject to terms and conditions
#  which differ from those listed here.
# - Neither the name of Analog Devices, Inc. nor the names of its contributors
#  may be used to endorse or promote products derived from this software
#  without specific prior written permission.
# - The use of this software may or may not infringe the patent rights of one
#  or more patent holders.  This license does not release you from the
#  requirement that you obtain separate licenses from these patent holders to
#  use this software.
#
# THIS SOFTWARE IS PROVIDED BY ANALOG DEVICES, INC. AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# NONINFRINGEMENT, TITLE, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL ANALOG DEVICES, INC. OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, PUNITIVE OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, DAMAGES ARISING OUT OF
# CLAIMS OF INTELLECTUAL PROPERTY RIGHTS INFRINGEMENT; PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# ******************************************************************************

import time
import logging
from datetime import datetime
from typing import Dict, List

from ..core.enums.pm_enums import LDO, ElectrodeSwitch, ControlStatus
from ..core.enums.dcb_enums import DCBCommand
from .common_application import CommonApplication
from ..core.enums.pm_enums import PMCommand, ChipID
from ..core.packets.command_packet import CommandPacket
from ..core.packets.common_packets import VersionPacket
from ..core.enums.common_enums import Application, CommonCommand
from ..core.packets.pm_packets import MCUVersionPacket, SystemInfoPacket, DateTimePacket, SyncTimerPacket, \
    HibernateModePacket, BootloaderVersionPacket, ToolAddressPacket
from ..core.packets.pm_packets import DCBStatusPacket, ChipIDPacket, UICRCustomerRegistersPacket, TopTouchControlPacket

logger = logging.getLogger(__name__)


class PMApplication(CommonApplication):
    """
    PM Application class.

    .. code-block:: python3
        :emphasize-lines: 4

        from adi_study_watch import SDK

        sdk = SDK("COM4")
        application = sdk.get_pm_application()
    """

    CHIP_ADPD4K = ChipID.ADPD4K
    CHIP_AD5940 = ChipID.AD5940
    CHIP_AD7156 = ChipID.AD7156
    CHIP_ADXL362 = ChipID.ADXL362
    CHIP_ADP5360 = ChipID.ADP5360
    CHIP_NAND_FLASH = ChipID.NAND_FLASH

    def __init__(self, packet_manager):
        super().__init__(Application.PM, packet_manager)
        self._args = {}
        self._csv_logger = {}
        self._last_timestamp = {}
        self._callback_function = {}
        self._dcb_size = 57

    def device_configuration_block_status(self) -> Dict:
        """
        Display dcb status of all applications.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_pm_application()
            x = application.device_configuration_block_status()
            print(x["payload"]["adxl_block"], x["payload"]["adpd4000_block"], x["payload"]["ppg_block"])
            # 0 0 0
        """
        request_packet = CommandPacket(self._destination, DCBCommand.QUERY_STATUS_REQ)
        response_packet = DCBStatusPacket(self._destination, DCBCommand.QUERY_STATUS_RES)
        return self._send_packet(request_packet, response_packet)

    def enter_boot_loader_mode(self) -> Dict:
        """
        Sets the device to boot loader mode.

        :return: A empty dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_pm_application()
            application.enter_boot_loader_mode()
        """
        request_packet = CommandPacket(self._destination, PMCommand.ENTER_BOOTLOADER_REQ)
        return self._send_packet_no_response(request_packet)

    def get_chip_id(self, chip_name: ChipID) -> Dict:
        """
        Returns chip id for specified chip name.

        :param chip_name: get chip id of the chip_name, use get_supported_chips() to list all support chip names.
        :type chip_name: ChipID
        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_pm_application()
            x = application.get_supported_chips()
            print(x)
            # [<ChipID.CHIP_ADXL362: ['0x1']>, ... , <ChipID.CHIP_AD7156: ['0x6']>]
            x = application.get_chip_id(application.CHIP_ADPD4K)
            print(x["payload"]["chip_name"], x["payload"]["chip_id"])
            # ChipID.ADPD4K 192
        """
        request_packet = ChipIDPacket(self._destination, PMCommand.CHIP_ID_REQ)
        request_packet.set_payload("chip_name", chip_name)
        response_packet = ChipIDPacket(self._destination, PMCommand.CHIP_ID_RES)
        return self._send_packet(request_packet, response_packet)

    def get_datetime(self) -> Dict:
        """
        Returns device current datetime.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_pm_application()
            x = application.get_datetime()
            print(f"{x['payload']['year']}-{x['payload']['month']}-{x['payload']['day']}")
            # 2020-12-16
            print(f"{x['payload']['hour']}:{x['payload']['minute']}:{x['payload']['second']}")
            # 15:17:57
        """
        request_packet = CommandPacket(self._destination, PMCommand.GET_DATE_TIME_REQ)
        response_packet = DateTimePacket(self._destination, PMCommand.GET_DATE_TIME_RES)
        return self._send_packet(request_packet, response_packet)

    def get_mcu_version(self) -> Dict:
        """
        Returns Device MCU version.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_pm_application()
            x = application.get_mcu_version()
            print(x["payload"]["mcu"])
            # MCUType.MCU_M4
        """
        request_packet = CommandPacket(self._destination, PMCommand.GET_MCU_VERSION_REQ)
        response_packet = MCUVersionPacket(self._destination, PMCommand.GET_MCU_VERSION_RES)
        return self._send_packet(request_packet, response_packet)

    @staticmethod
    def get_supported_chips() -> List[ChipID]:
        """
        List all supported chips for PM.

        :return: Array of chips enums.
        :rtype: List

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_pm_application()
            x = application.get_supported_chips()
            print(x)
            # [<ChipID.CHIP_ADXL362: ['0x1']>, ... , <ChipID.CHIP_AD7156: ['0x6']>]
        """
        return [PMApplication.CHIP_ADXL362, PMApplication.CHIP_ADPD4K, PMApplication.CHIP_ADP5360,
                PMApplication.CHIP_AD5940, PMApplication.CHIP_NAND_FLASH, PMApplication.CHIP_AD7156]

    def get_system_info(self) -> Dict:
        """
        Returns Device system info.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_pm_application()
            x = application.get_system_info()
            print(x["payload"]["version"])
            # 0
            print(x["payload"]["mac_address"])
            # C5-05-CA-F1-67-D5
            print(x["payload"]["device_id"])
            # 0
        """
        request_packet = CommandPacket(self._destination, PMCommand.SYS_INFO_REQ)
        response_packet = SystemInfoPacket(self._destination, PMCommand.SYS_INFO_RES)
        return self._send_packet(request_packet, response_packet)

    def enable_battery_charging(self) -> Dict:
        """
        Enable battery charging.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_pm_application()
            application.enable_battery_charging()
        """
        request_packet = CommandPacket(self._destination, PMCommand.ENABLE_BAT_CHARGE_REQ)
        response_packet = CommandPacket(self._destination, PMCommand.ENABLE_BAT_CHARGE_RES)
        return self._send_packet(request_packet, response_packet)

    def disable_battery_charging(self) -> Dict:
        """
        Disable battery charging.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_pm_application()
            application.disable_battery_charging()
        """
        request_packet = CommandPacket(self._destination, PMCommand.DISABLE_BAT_CHARGE_REQ)
        response_packet = CommandPacket(self._destination, PMCommand.DISABLE_BAT_CHARGE_RES)
        return self._send_packet(request_packet, response_packet)

    def get_version(self) -> Dict:
        """
        Returns Device version info.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_pm_application()
            x = application.get_version()
            print(x["payload"]["major_version"])
            # 1
            print(x["payload"]["minor_version"])
            # 0
            print(x["payload"]["patch_version"])
            # 1
            print(x["payload"]["version_string"])
            # -Perseus
            print(x["payload"]["build_version"])
            # |298b4ce1_Rl|2020-12-14 12:34:31 -0500
        """
        request_packet = CommandPacket(self._destination, CommonCommand.GET_VERSION_REQ)
        response_packet = VersionPacket(self._destination, CommonCommand.GET_VERSION_RES)
        return self._send_packet(request_packet, response_packet)

    def set_datetime(self, date_time: datetime) -> Dict:
        """
        Set specified datetime to device.

        :param date_time: datetime for device.
        :type date_time: datetime
        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK
            import datetime

            sdk = SDK("COM4")
            application = sdk.get_pm_application()
            now = datetime.datetime.now()
            x = application.set_datetime(now)
            print(x["payload"]["status"])
            # CommonStatus.OK
        """
        is_dst = time.daylight and time.localtime().tm_isdst > 0
        utc_offset = - (time.altzone if is_dst else time.timezone)
        request_packet = DateTimePacket(self._destination, PMCommand.SET_DATE_TIME_REQ)
        request_packet.set_payload("year", date_time.year)
        request_packet.set_payload("month", date_time.month)
        request_packet.set_payload("day", date_time.day)
        request_packet.set_payload("hour", date_time.hour)
        request_packet.set_payload("minute", date_time.minute)
        request_packet.set_payload("second", date_time.second)
        request_packet.set_payload("tz_sec", utc_offset)
        response_packet = CommandPacket(self._destination, PMCommand.SET_DATE_TIME_RES)
        return self._send_packet(request_packet, response_packet)

    def system_hardware_reset(self) -> Dict:
        """
        Reset device hardware.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_pm_application()
            x = application.system_hardware_reset()
            print(x["payload"]["status"])
            # PMStatus.OK
        """
        request_packet = CommandPacket(self._destination, PMCommand.SYSTEM_HW_RESET_REQ)
        response_packet = CommandPacket(self._destination, PMCommand.SYSTEM_HW_RESET_RES)
        return self._send_packet(request_packet, response_packet)

    def system_reset(self) -> Dict:
        """
        Reset device system.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_pm_application()
            x = application.system_reset()
            print(x["payload"]["status"])
            # PMStatus.OK
        """
        request_packet = CommandPacket(self._destination, PMCommand.SYSTEM_RESET_REQ)
        response_packet = CommandPacket(self._destination, PMCommand.SYSTEM_RESET_RES)
        return self._send_packet(request_packet, response_packet)

    def read_uicr_customer_registers(self) -> Dict:
        """
        Read UICR customer Register.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_pm_application()
            packet = application.read_uicr_customer_registers()

        """
        request_packet = UICRCustomerRegistersPacket(self._destination, PMCommand.READ_UICR_CUSTOMER_REG_REQ)
        response_packet = UICRCustomerRegistersPacket(self._destination, PMCommand.READ_UICR_CUSTOMER_REG_RES)
        return self._send_packet(request_packet, response_packet)

    def write_uicr_customer_registers(self, hw_revision: str = None, manufacture_date: str = None) -> Dict:
        """
        Writes to UICR customer Register.
        This API can only be used after NOR flash is erased, the following command can be used to do so:
        $nrfjprog --eraseall

        The following information is written to UICR. Hardcoded information:
            manufacturer_name "ADI",
            model_number "EVAL-HCRWATCH4Z
            hw_revision "4.3.00"
            serial_number 12 bytes BLE MAC address
            manufacture_date is current system date manufacture_date This is optional.

        :param manufacture_date: manufacture date size limit 12. default = current system date(YYYY-MM-DD).
        :param hw_revision: hw revision size limit 8. default = 4.3.00.
        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_pm_application()
            packet = application.write_uicr_customer_registers("2022-02-15")

        """
        mac_address = self.get_system_info()["payload"]["mac_address"]
        now = datetime.now()
        date = str(now.year) + "-" + str('%02d' % now.month) + "-" + str('%02d' % now.day)
        if manufacture_date is None:
            manufacture_date = date
        if hw_revision is None:
            hw_revision = "4.3.00"
        request_packet = UICRCustomerRegistersPacket(self._destination, PMCommand.WRITE_UICR_CUSTOMER_REG_REQ)
        request_packet.set_payload("manufacturer_name", "ADI")
        request_packet.set_payload("model_number", "EVAL-HCRWATCH4Z")
        request_packet.set_payload("hw_revision", hw_revision)
        request_packet.set_payload("serial_number", mac_address.replace("-", ""))
        request_packet.set_payload("manufacture_date", manufacture_date)
        request_packet.set_payload("crc_8", 0)
        response_packet = UICRCustomerRegistersPacket(self._destination, PMCommand.WRITE_UICR_CUSTOMER_REG_RES)
        return self._send_packet(request_packet, response_packet)

    def enable_sync_timer(self) -> Dict:
        """
        Enables Sync Timer

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_pm_application()
            application.enable_sync_timer()
        """
        request_packet = SyncTimerPacket(self._destination, PMCommand.SYNC_TIMER_ENABLE_REQ)
        request_packet.set_payload("enabled", 1)
        response_packet = SyncTimerPacket(self._destination, PMCommand.SYNC_TIMER_ENABLE_RES)
        return self._send_packet(request_packet, response_packet)

    def disable_sync_timer(self) -> Dict:
        """
        Disable Sync Timer

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_pm_application()
            application.disable_sync_timer()
        """
        request_packet = SyncTimerPacket(self._destination, PMCommand.SYNC_TIMER_ENABLE_REQ)
        request_packet.set_payload("enabled", 0)
        response_packet = SyncTimerPacket(self._destination, PMCommand.SYNC_TIMER_ENABLE_RES)
        return self._send_packet(request_packet, response_packet)

    def start_sync_timer(self) -> Dict:
        """
        Starts Sync Timer

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_pm_application()
            application.start_sync_timer()
        """
        request_packet = SyncTimerPacket(self._destination, PMCommand.SYNC_TIMER_START_STOP_REQ)
        request_packet.set_payload("enabled", 1)
        response_packet = SyncTimerPacket(self._destination, PMCommand.SYNC_TIMER_START_STOP_RES)
        return self._send_packet(request_packet, response_packet)

    def stop_sync_timer(self) -> Dict:
        """
        Stops Sync Timer

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_pm_application()
            application.enable_sync_timer()
        """
        request_packet = SyncTimerPacket(self._destination, PMCommand.SYNC_TIMER_START_STOP_REQ)
        request_packet.set_payload("enabled", 0)
        response_packet = SyncTimerPacket(self._destination, PMCommand.SYNC_TIMER_START_STOP_RES)
        return self._send_packet(request_packet, response_packet)

    def enable_top_touch(self) -> Dict:
        """
        This API is used to enable top touch application(backlight control) in Fw.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_pm_application()
            application.enable_top_touch()
        """
        request_packet = TopTouchControlPacket(self._destination, PMCommand.SET_TOP_TOUCH_CONTROL_REQ)
        request_packet.set_payload("enabled", 1)
        response_packet = TopTouchControlPacket(self._destination, PMCommand.SET_TOP_TOUCH_CONTROL_RES)
        return self._send_packet(request_packet, response_packet)

    def disable_top_touch(self) -> Dict:
        """
        This API is used to disable top touch application(backlight control) in Fw.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_pm_application()
            application.disable_top_touch()
        """
        request_packet = TopTouchControlPacket(self._destination, PMCommand.SET_TOP_TOUCH_CONTROL_REQ)
        request_packet.set_payload("enabled", 0)
        response_packet = TopTouchControlPacket(self._destination, PMCommand.SET_TOP_TOUCH_CONTROL_RES)
        return self._send_packet(request_packet, response_packet)

    def get_top_touch_status(self) -> Dict:
        """
        This API is used to get the status of top touch application(backlight control) in Fw,
         whether it is enabled or disabled

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_pm_application()
            application.get_top_touch_status()
        """
        request_packet = TopTouchControlPacket(self._destination, PMCommand.GET_TOP_TOUCH_CONTROL_REQ)
        request_packet.set_payload("enabled", 0)
        response_packet = TopTouchControlPacket(self._destination, PMCommand.GET_TOP_TOUCH_CONTROL_RES)
        return self._send_packet(request_packet, response_packet)

    def load_configuration(self) -> Dict:
        """
        This command must be used every time after PM DCB write/erase to update the device
        configuration. Otherwise not needed as ADP5360 config is already loaded during boot up.


        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_pm_application()
            application.load_configuration()
        """
        request_packet = CommandPacket(self._destination, PMCommand.LOAD_CFG_REQ)
        response_packet = CommandPacket(self._destination, PMCommand.LOAD_CFG_RES)
        return self._send_packet(request_packet, response_packet)

    def fds_erase(self) -> Dict:
        """
        Erases the NOR flash pages reserved for FDS;
        This command will result in BLE disconnection and watch reset.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_pm_application()
            application.fds_erase()
        """
        request_packet = CommandPacket(self._destination, PMCommand.FDS_ERASE_REQ)
        response_packet = CommandPacket(self._destination, PMCommand.FDS_ERASE_RES)
        return self._send_packet(request_packet, response_packet)

    def enable_hibernate_mode(self, seconds_to_trigger: int) -> Dict:
        """
        Enable the hibernate mode status in the Watch Fw.
        It is based on this value that Hibernate Mode happens.

        :param seconds_to_trigger: Time in seconds, based on which the Hibernate mode entry happens.
        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_pm_application()
            application.enable_hibernate_mode()
        """
        request_packet = HibernateModePacket(self._destination, PMCommand.SET_HIBERNATE_MODE_STATUS_REQ)
        request_packet.set_payload("hibernate_mode", 1)
        request_packet.set_payload("seconds_to_trigger", seconds_to_trigger)
        response_packet = HibernateModePacket(self._destination, PMCommand.SET_HIBERNATE_MODE_STATUS_RES)
        return self._send_packet(request_packet, response_packet)

    def disable_hibernate_mode(self) -> Dict:
        """
        Disable the hibernate mode status in the Watch Fw.
        It is based on this value that Hibernate Mode happens.
        It also shows the time in seconds, based on which the Hibernate mode entry happens.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_pm_application()
            application.disable_hibernate_mode()
        """
        request_packet = HibernateModePacket(self._destination, PMCommand.SET_HIBERNATE_MODE_STATUS_REQ)
        request_packet.set_payload("hibernate_mode", 0)
        request_packet.set_payload("seconds_to_trigger", 0xFFFF)
        response_packet = HibernateModePacket(self._destination, PMCommand.SET_HIBERNATE_MODE_STATUS_RES)
        return self._send_packet(request_packet, response_packet)

    def get_hibernate_mode(self) -> Dict:
        """
        Get the hibernate mode status- enabled/disabled in the Watch Fw.
        It is based on this value that Hibernate Mode happens.
        It also shows the time in seconds, based on which the Hibernate mode entry happens.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_pm_application()
            application.get_hibernate_mode()
        """
        request_packet = CommandPacket(self._destination, PMCommand.GET_HIBERNATE_MODE_STATUS_REQ)
        response_packet = HibernateModePacket(self._destination, PMCommand.GET_HIBERNATE_MODE_STATUS_RES)
        return self._send_packet(request_packet, response_packet)

    def get_bootloader_version(self) -> Dict:
        """
        Get the bootloader's version information, if Bootloader is present in the Watch.
        This returns 0, if Bootloader is not present or if Bootloader DFU has not yet been done.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_pm_application()
            application.get_bootloader_version()
        """
        request_packet = CommandPacket(self._destination, PMCommand.GET_BOOTLOADER_VERSION_REQ)
        response_packet = BootloaderVersionPacket(self._destination, PMCommand.GET_BOOTLOADER_VERSION_RES)
        return self._send_packet(request_packet, response_packet)

    def get_existing_connection(self):
        """
        Connect to serial device over USB or BLE, after checking if there is previous tool staying connected to the watch.
        Provide a serial device identifier appropriate for your platform (COMX for Windows, /dev/ttyX for Linux and OSX).

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_pm_application()
            application.connect_usb_check_previous_tool_connection()
        """
        request_packet = CommandPacket(self._destination, PMCommand.GET_CONNECTED_TOOL_ADDR_REQ)
        response_packet = ToolAddressPacket(self._destination, PMCommand.GET_CONNECTED_TOOL_ADDR_RES)
        return self._send_packet(request_packet, response_packet)



# ******************************************************************************
# Copyright (c) 2019 Analog Devices, Inc.  All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# - Redistributions of source code must retain the above copyright notice, this
#  list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice,
#  this list of conditions and the following disclaimer in the documentation
#  and/or other materials provided with the distribution.
# - Modified versions of the software must be conspicuously marked as such.
# - This software is licensed solely and exclusively for use with
#  processors/products manufactured by or for Analog Devices, Inc.
# - This software may not be combined or merged with other code in any manner
#  that would cause the software to become subject to terms and conditions
#  which differ from those listed here.
# - Neither the name of Analog Devices, Inc. nor the names of its contributors
#  may be used to endorse or promote products derived from this software
#  without specific prior written permission.
# - The use of this software may or may not infringe the patent rights of one
#  or more patent holders.  This license does not release you from the
#  requirement that you obtain separate licenses from these patent holders to
#  use this software.
#
# THIS SOFTWARE IS PROVIDED BY ANALOG DEVICES, INC. AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# NONINFRINGEMENT, TITLE, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL ANALOG DEVICES, INC. OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, PUNITIVE OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, DAMAGES ARISING OUT OF
# CLAIMS OF INTELLECTUAL PROPERTY RIGHTS INFRINGEMENT; PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# ******************************************************************************

from unittest import TestCase

from sdk_pure_python.adi_study_watch import SDK
from sdk_pure_python.adi_study_watch.core.enums.common_enums import CommonStatus
from sdk_pure_python.adi_study_watch.core.enums.fs_enums import FSStatus


class TestFSApplication(TestCase):

    @classmethod
    def setUpClass(cls):
        cls.application = SDK("COM4", debug=True).get_fs_application()

    def test_abort_logging(self):
        x = self.application.abort_logging()
        assert (x["payload"]["status"] == CommonStatus.STREAM_NOT_STARTED)

    def test_config_log(self):
        self.application.disable_config_log()
        self.application.enable_config_log()

    def test_delete_config_file(self):
        self.application.delete_config_file()

    def test_logging(self):
        self.application.start_logging()
        self.application.stop_logging()

    def test_get_file_count(self):
        x = self.application.get_file_count()
        assert (x["payload"]["status"] == CommonStatus.OK)

    def test_format(self):
        x = self.application.format()
        assert (x["payload"]["status"] == FSStatus.OK)

    def test_get_stream_status(self):
        x = self.application.get_stream_status(self.application.STREAM_ADXL)
        assert (x["payload"]["status"] == FSStatus.OK)

    # def test_inject_key_value_pair(self):
    #     x = self.application.inject_key_value_pair(1234)
    #     assert (x["payload"]["status"] == FSStatus.OK)

    # def test_ls(self):
    #     x = self.application.ls()
    #     assert (type(x) == list)

    # def test_mount(self):
    #     x = self.application.mount()
    #     assert (x["payload"]["status"] == FSStatus.OK)

    def test_get_status(self):
        x = self.application.get_status()
        assert (x["payload"]["status"] == FSStatus.LOGGING_NOT_STARTED)

    # def test_subscribe_stream(self):
    #     x = self.application.subscribe_stream(self.application.STREAM_ADXL)
    #     assert (x["payload"]["status"] == CommonStatus.SUBSCRIBER_ADDED)
    #     x = self.application.unsubscribe_stream(self.application.STREAM_ADXL)
    #     assert (x["payload"]["status"] == CommonStatus.SUBSCRIBER_REMOVED)

    def test_volume_info(self):
        x = self.application.volume_info()
        assert (x["payload"]["status"] == FSStatus.OK)



# ******************************************************************************
# Copyright (c) 2019 Analog Devices, Inc.  All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# - Redistributions of source code must retain the above copyright notice, this
#  list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice,
#  this list of conditions and the following disclaimer in the documentation
#  and/or other materials provided with the distribution.
# - Modified versions of the software must be conspicuously marked as such.
# - This software is licensed solely and exclusively for use with
#  processors/products manufactured by or for Analog Devices, Inc.
# - This software may not be combined or merged with other code in any manner
#  that would cause the software to become subject to terms and conditions
#  which differ from those listed here.
# - Neither the name of Analog Devices, Inc. nor the names of its contributors
#  may be used to endorse or promote products derived from this software
#  without specific prior written permission.
# - The use of this software may or may not infringe the patent rights of one
#  or more patent holders.  This license does not release you from the
#  requirement that you obtain separate licenses from these patent holders to
#  use this software.
#
# THIS SOFTWARE IS PROVIDED BY ANALOG DEVICES, INC. AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# NONINFRINGEMENT, TITLE, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL ANALOG DEVICES, INC. OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, PUNITIVE OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, DAMAGES ARISING OUT OF
# CLAIMS OF INTELLECTUAL PROPERTY RIGHTS INFRINGEMENT; PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# ******************************************************************************
from typing import List, Dict
from datetime import datetime, timezone, timedelta

from ..core.enums.pm_enums import PMCommand
from .common_application import CommonApplication
from ..core.enums.common_enums import Application
from ..core.packets.pm_packets import DateTimePacket
from ..core.packets.command_packet import CommandPacket
from ..core.enums.lt_mode4_enums import LTMode4State, LTMode4Command
from ..core.packets.lt_mode4_packets import LTMode4PrevStateEventPacket, LTMode4StatePacket


class LTMode4Application(CommonApplication):
    """
    lt_mode4 Application class.

    .. code-block:: python3
        :emphasize-lines: 4

        from adi_study_watch import SDK

        sdk = SDK("COM4")
        application = sdk.get_lt_mode4_application()

    """

    STANDBY = LTMode4State.STANDBY
    CONFIGURED = LTMode4State.CONFIGURED
    LOG_ON = LTMode4State.LOG_ON
    LOG_OFF = LTMode4State.LOG_OFF
    FS_MEMORY_FULL = LTMode4State.FS_MEMORY_FULL
    FS_MAX_FILE_COUNT = LTMode4State.FS_MAX_FILE_COUNT
    SHIPMENT_MODE = LTMode4State.SHIPMENT_MODE
    LOG_OFF_TOOL_CONNECTED = LTMode4State.LOG_OFF_TOOL_CONNECTED
    LOG_DOWNLOAD = LTMode4State.LOG_DOWNLOAD

    def __init__(self, packet_manager):
        super().__init__(Application.LT_MODE4_APP, packet_manager)
        self._dcb_size = 19

    @staticmethod
    def get_supported_states():
        """
        List all supported states for LTMode4State.

        :return: Array of lt_mode4 state ID enums.
        :rtype: List[LTMode4State]
        """
        return [enum_value for enum_value in LTMode4State]

    def clear_prev_state_event(self):
        """
         Clears the LTMode4 config app's previous state, event received and the corresponding timestamp structure
         maintained registered in the Watch Fw.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_lt_mode4_application()
            application.clear_prev_state_event()

        """
        request_packet = CommandPacket(self._destination, LTMode4Command.CLEAR_PREV_ST_EVT_REQ)
        response_packet = CommandPacket(self._destination, LTMode4Command.CLEAR_PREV_ST_EVT_RES)
        return self._send_packet(request_packet, response_packet)

    def get_prev_state_event(self):
        """
        This is a command, to get the lt_mode4 config app's previous state, event received and the corresponding timestamp
        registered in the Watch Fw.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_lt_mode4_application()
            application.get_prev_state_event()

        """
        # TODO: FIX THIS
        request_packet = CommandPacket(Application.PM, PMCommand.GET_DATE_TIME_REQ)
        response_packet = DateTimePacket(Application.PM, PMCommand.GET_DATE_TIME_RES)
        date_time_packet = self._send_packet(request_packet, response_packet)
        offset = timezone(timedelta(seconds=date_time_packet["payload"]["tz_sec"]))
        date_time = datetime(date_time_packet['payload']['year'], date_time_packet['payload']['month'],
                             date_time_packet['payload']['day'], date_time_packet['payload']['hour'],
                             date_time_packet['payload']['minute'], date_time_packet['payload']['second'],
                             tzinfo=offset)

        request_packet = CommandPacket(self._destination, LTMode4Command.GET_PREV_ST_EVT_REQ)
        response_packet = LTMode4PrevStateEventPacket(self._destination, LTMode4Command.GET_PREV_ST_EVT_RES)
        packet = self._send_packet(request_packet, response_packet)
        ts = (32000.0 * ((date_time.hour * 3600) + (date_time.minute * 60) + date_time.second))
        ref_time = date_time.timestamp()
        for i in range(len(packet["payload"]["stream_data"])):
            timestamp = packet["payload"]["stream_data"][i]["timestamp"]
            change = ts - timestamp
            change = change / 32000.0
            reference_time = ref_time - change
            packet["payload"]["stream_data"][i]["timestamp"] = reference_time * 1000
        return packet

    def get_state(self):
        """
        Get lt_mode4 state.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_lt_mode4_application()
            application.get_state()

        """
        request_packet = CommandPacket(self._destination, LTMode4Command.GET_STATE_REQ)
        response_packet = LTMode4StatePacket(self._destination, LTMode4Command.GET_STATE_RES)
        return self._send_packet(request_packet, response_packet)

    def set_state(self, state: LTMode4State):
        """
        Set lt_mode4 state.

        :param state: lt_mode4 state, use get_supported_states() to list all lt_mode4 states.
        :type state: LTMode4State
        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_lt_mode4_application()
            application.set_state(application.STATE_SLEEP)

        """
        request_packet = LTMode4StatePacket(self._destination, LTMode4Command.SET_STATE_REQ)
        request_packet.set_payload("state", state)
        response_packet = LTMode4StatePacket(self._destination, LTMode4Command.SET_STATE_RES)
        return self._send_packet(request_packet, response_packet)
